/*
    This file is part of darktable,
    copyright (c) 2009--2010 johannes hanika.

    darktable is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    darktable is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with darktable.  If not, see <http://www.gnu.org/licenses/>.
*/
/* config.h.  Generated from config.h.in by configure.  */
/* config.h.in.  Generated from configure.ac by autoheader.  */

/* Define to 1 if you are compiling for PowerPC. */
/* #undef ARCH_PPC */

/* Define to 1 if you are compiling for PowerPC64. */
/* #undef ARCH_PPC64 */

/* Define to 1 if you are compiling for ix86. */
#define ARCH_X86 1

/* Define to 1 if you are compiling for amd64. */
/* #undef ARCH_X86_64 */

/* Dynamic shared library */
#define DYNAMICLIB "-shared"

/* The name of the GEGL library */
#define GEGL_LIBRARY "gegl-0.0"

/* Define to 1 if this is an unstable version of GEGL */
/* #undef GEGL_UNSTABLE */

/* Our gettext translation domain. */
#define GETTEXT_PACKAGE "gegl-0.0"

/* Define to 1 if you have the `bind_textdomain_codeset' function. */
#define HAVE_BIND_TEXTDOMAIN_CODESET 1

/* Define to 1 if you have the <dlfcn.h> header file. */
#define HAVE_DLFCN_H 1

/* Define to 1 to compile with gio support. */
#define HAVE_GIO 1

/* Define to 1 to compile with gtk support. */
#define HAVE_GTK 1

/* Define to 1 if you have the <inttypes.h> header file. */
#define HAVE_INTTYPES_H 1

/* define to 1 if avformat.h is found in libavformat */
/* #undef HAVE_LIBAVFORMAT_AVFORMAT_H */

/* Define to 1 if you have the <memory.h> header file. */
#define HAVE_MEMORY_H 1

/* Define to 1 if the spiro library is available */
/* #undef HAVE_SPIRO */

/* Define to 1 if you have the <stdint.h> header file. */
#define HAVE_STDINT_H 1

/* Define to 1 if you have the <stdlib.h> header file. */
#define HAVE_STDLIB_H 1

/* Define to 1 if you have the <strings.h> header file. */
#define HAVE_STRINGS_H 1

/* Define to 1 if you have the <string.h> header file. */
#define HAVE_STRING_H 1

/* Define to 1 if you have the <sys/stat.h> header file. */
#define HAVE_SYS_STAT_H 1

/* Define to 1 if you have the <sys/types.h> header file. */
#define HAVE_SYS_TYPES_H 1

/* Define to 1 if you have the <unistd.h> header file. */
#define HAVE_UNISTD_H 1

/* Define to the sub-directory in which libtool stores uninstalled libraries.
   */
#define LT_OBJDIR ".libs/"

/* Define to the address where bug reports for this package should be sent. */
#define PACKAGE_BUGREPORT ""

/* Define to the full name of this package. */
#define PACKAGE_NAME "gegl"

/* Define to the full name and version of this package. */
#define PACKAGE_STRING "gegl 0.0.22"

/* Define to the one symbol short name of this package. */
#define PACKAGE_TARNAME "gegl"

/* Define to the version of this package. */
#define PACKAGE_VERSION "0.0.22"

/* File extension for shared libraries */
#define SHREXT ".so"

/* Define to 1 if you have the ANSI C header files. */
#define STDC_HEADERS 1

/* Define to 1 if MMX assembly is available. */
#define USE_MMX 1

/* Define to 1 if SSE assembly is available. */
#define USE_SSE 1

/* Number of bits in a file offset, on hosts where this is settable. */
#define _FILE_OFFSET_BITS 64

/* Define for large files, on AIX-style hosts. */
/* #undef _LARGE_FILES */
