#ifndef BLACK_AREA_H
#define BLACK_AREA_H

/* 
    RawSpeed - RAW file decoder.

    Copyright (C) 2009 Klaus Post

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

    http://www.klauspost.com
*/
namespace RawSpeed {

class BlackArea
{
public:
  BlackArea(int offset, int size, bool isVertical);
  virtual ~BlackArea(void);
  uint32 offset; // Offset in bayer pixels.
  uint32 size;   // Size in bayer pixels.
  bool isVertical;  // Otherwise horizontal
};

} // namespace RawSpeed

#endif
