/*
    This file is part of darktable,
    copyright (c) 2011 tobias ellinghaus.

    darktable is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    darktable is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with darktable.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __GROUPING_H__
#define __GROUPING_H__

/** add an image to a group */
void dt_grouping_add_to_group(int group_id, int image_id);

/** remove an image from a group. returns the new group_id of the other images. */
int dt_grouping_remove_from_group(int image_id);

/** make an image the representative of the group it is in. returns the new group_id. */
int dt_grouping_change_representative(int image_id);

#endif

// kate: tab-indents: off; indent-width 2; replace-tabs on; indent-mode cstyle; remove-trailing-space on;
