#ifndef _OSM_GPS_MAP_H_
#define _OSM_GPS_MAP_H_

#include <osm-gps-map-osd.h>
#include <osm-gps-map-layer.h>
#include <osm-gps-map-track.h>
#include <osm-gps-map-point.h>
#include <osm-gps-map-image.h>
#include <osm-gps-map-source.h>
#include <osm-gps-map-widget.h>
#include <osm-gps-map-compat.h>

#endif /* _OSM_GPS_MAP_H_ */

